/* func es6 */
/* global exports module define require */
/* eslint new-cap: 0, no-multi-spaces: 0 */
/**
 * Converts SmartType to a JavaScript function
 */
(function () {
    "use strict";

    var factory = function (smarttype, simplifier, P)
    {
        var parser = smarttype.parser;
        var Simplifier = simplifier.Simplifier;

        var UnsupportedMathError = function (token)
        {
            this.name = "UnsupportedMathError";
            this.message = "Unsupported math: " + token;
        };
        UnsupportedMathError.prototype = new Error();
        UnsupportedMathError.prototype.constructor = UnsupportedMathError;

        var Visitor = function (options)
        {
            this.options = options;
        };
        var add_f = function (operands, bindings)
        {
            var r = operands[0](bindings);
            for (var i = 1; i < operands.length - 1; i += 2)
            {
                if (operands[i] === '+')
                {
                    r += operands[i+1](bindings);
                }
                else
                {
                    r -= operands[i+1](bindings);
                }
            }
            return r;
        };
        Visitor.prototype.visit_addition = function ()
        {
            return add_f.bind(undefined, Array.prototype.slice.call(arguments, 1));
        };
        Visitor.prototype.visit_addition_sign = function (sign)
        {
            if (sign[0] === '\\')
            {
                throw new UnsupportedMathError(sign);
            }
            return sign;
        };
        var multiply_f = function (operands, bindings)
        {
            var r = 1;
            for (var i = 0; i < operands.length; i += 2)
            {
                r *= operands[i](bindings);
            }
            return r;
        };
        Visitor.prototype.visit_multiplication = function ()
        {
            return multiply_f.bind(undefined, Array.prototype.slice.call(arguments, 1));
        };
        Visitor.prototype.visit_multiplication_op = function (visible)
        {
            return;
        };

        var reciprocal = function (x)
        {
            return 1 / x;
        };
        var div = function (x, y)
        {
            return x / y;
        };
        var id = function (n)
        {
            return n;
        };
        var neg = function (x)
        {
            return -x;
        };

        var const_f = function (c, bindings)
        {
            return c;
        };
        var apply1_f = function (f, operand, bindings)
        {
            return f.call(undefined, operand(bindings));
        };
        var apply2_f = function (f, op1, op2, bindings)
        {
            return f.call(undefined, op1(bindings), op2(bindings));
        };
        var log_f = function (base, operand, bindings)
        {
            return Math.log(operand)/Math.log(base(bindings));
        };

        Visitor.prototype.visit_binop = new P.Function()
            .when(P._, P('left'), 'root',  P('right'), _ => apply2_f.bind(undefined, Math.pow, _.right, apply1_f.bind(undefined, reciprocal, _.left)))
            .when(P._, P('left'), '^',     P('right'), _ => apply2_f.bind(undefined, Math.pow, _.left, _.right))
            .when(P._, P('left'), '/',     P('right'), _ => apply2_f.bind(undefined, div, _.left, _.right))
            .when(P._, P('left'), '=',     P('right'), _ => ({left: _.left, op: '=', right: _.right}))
            .when(P._, 'log',     '_',     P('base'),  _ => log_f.bind(undefined, _.base))
            .when(P._, P('left', P.isA(Array)), ',', P('right'), function (_) { _.left.push(_.right); return _.left; })
            .when(P._, P('left'), ',',     P('right'), _ => [_.left, _.right])
            .when(P._, P._,       P('op'), P._,        _ => { throw new UnsupportedMathError(_.op); })
            .end;
        Visitor.prototype.visit_funcname = function (name)
        {
            return name;
        };
        var sum_f = function (variable, start, end, val, bindings)
        {
            var newbindings = {};
            if (bindings)
            {
                for (var k in bindings)
                {
                    newbindings[k] = bindings[k];
                }
            }
            variable = variable(variable_f);
            start = start(bindings);
            end = end(bindings);
            var r = 0;
            for (var i = start; i <= end; i++)
            {
                newbindings[variable] = i;
                r += val(newbindings);
            }
            return r;
        };
        var prod_f = function (variable, start, end, val, bindings)
        {
            var newbindings = {};
            if (bindings)
            {
                for (var k in bindings)
                {
                    newbindings[k] = bindings[k];
                }
            }
            variable = variable(variable_f);
            start = start(bindings);
            end = end(bindings);
            var r = 1;
            for (var i = start; i <= end; i++)
            {
                newbindings[variable] = i;
                r *= val(newbindings);
            }
            return r;
        };
        Visitor.prototype.visit_bigop = new P.Function()
            .when(P._, 'sum',   {left: P('variable'),
                                 op: '=',
                                 right: P('start')}, P('end'), _ => ({func: 'sum', variable: _.variable, start: _.start, end: _.end}))
            .when(P._, 'prod',  {left: P('variable'),
                                 op: '=',
                                 right: P('start')}, P('end'), _ => ({func: 'prod', variable: _.variable, start: _.start, end: _.end}))
            .when(P._, P('op'), P('sub'), P('sup'), _ => { throw new UnsupportedMathError(_.op); })
            .end;
        Visitor.prototype.visit_sign = new P.Function()
            .when(P._, '+',     P('value'), _ => _.value)
            .when(P._, '-',     P('value'), _ => apply1_f.bind(undefined, neg, _.value))
            .when(P._, P('op'), P._,        _ => { throw new UnsupportedMathError(_.op); })
            .end;
        var min_f = function (values, bindings)
        {
            var v = [];
            for (var i = 0; i < values.length; i++)
            {
                v.push(values[i](bindings));
            }
            return Math.min.apply(undefined, v);
        };
        var max_f = function (values, bindings)
        {
            var v = [];
            for (var i = 0; i < values.length; i++)
            {
                v.push(values[i](bindings));
            }
            return Math.max.apply(undefined, v);
        };
        Visitor.prototype.visit_func = new P.Function()
        // special functions
            .when(P._, {func: 'sum',
                        variable: P('variable'),
                        start: P('start'),
                        end: P('end')}, P('value'), _ => sum_f.bind(undefined, _.variable,  _.start, _.end, _.value))
            .when(P._, {func: 'prod',
                        variable: P('variable'),
                        start: P('start'),
                        end: P('end')}, P('value'), _ => prod_f.bind(undefined, _.variable,  _.start, _.end, _.value))
            .when(P._, P('func', P.isA(Function)), P('value'), _ => apply1_f.bind(undefined, _.func, _.value))
            .when(P._, 'min',      P('values', P.isAn(Array)), _ => min_f.bind(undefined, _.values))
            .when(P._, 'min',      P('value'), _ => _.value)
            .when(P._, 'max',      P('values', P.isAn(Array)), _ => max_f.bind(undefined, _.values))
            .when(P._, 'max',      P('value'), _ => _.value)
        // regular functions
            .when(P._, 'floor',    P('value'), _ => apply1_f.bind(undefined, Math.floor, _.value))
            .when(P._, 'ceil',     P('value'), _ => apply1_f.bind(undefined, Math.ceil, _.value))
            .when(P._, 'sqrt',     P('value'), _ => apply1_f.bind(undefined, Math.sqrt, _.value))
            .when(P._, 'abs',      P('value'), _ => apply1_f.bind(undefined, Math.abs, _.value))
            .when(P._, 'ln',       P('value'), _ => apply1_f.bind(undefined, Math.log, _.value))
            .when(P._, 'sin',      P('value'), _ => apply1_f.bind(undefined, Math.sin, _.value))
            .when(P._, 'cos',      P('value'), _ => apply1_f.bind(undefined, Math.cos, _.value))
            .when(P._, 'tan',      P('value'), _ => apply1_f.bind(undefined, Math.tan, _.value))
            .when(P._, 'arcsin',   P('value'), _ => apply1_f.bind(undefined, Math.asin, _.value))
            .when(P._, 'arccos',   P('value'), _ => apply1_f.bind(undefined, Math.acos, _.value))
            .when(P._, 'arctan',   P('value'), _ => apply1_f.bind(undefined, Math.atan, _.value))
            .when(P._, P('name'),  P('value'), _ => { throw new UnsupportedMathError(_.name); })
            .end;
        Visitor.prototype.visit_postfunc = new P.Function()
        // FIXME: support factorial
            .when(P._, P('name'),  P('value'), _ => { throw new UnsupportedMathError(_.name); })
            .end;
        Visitor.prototype.visit_number = function (num)
        {
            return id.bind(undefined, parseFloat(num.value));
        };
        Visitor.prototype.visit_input = function (input)
        {//TODO:
            throw new Error('Not Supported');
            return '';
        };
        Visitor.prototype.visit_latex = function (latex)
        {//TODO:
            throw new Error('Not Supported');
            return '';
        };
        var variable_f = function (name, bindings)
        {
            if (bindings === variable_f)
            {
                return name;
            }
            return bindings[name];
        };
        Visitor.prototype.visit_variable = new P.Function()
            .when({name: 'e'},       _ => const_f.bind(undefined, Math.E))
            // FIXME: make sure variable is available
            .when({name: P('name')}, _ => variable_f.bind(undefined, _.name))
            .end;
        var greek = /^(?:alpha|beta|[Gg]amma|[Dd]elta|(?:var)?epsilon|(?:z||(?:T|(?:var)?t)h)eta|iota|(?:var)?kappa|[Ll]ambda|[mn]u|[Xx]i|omicron|(?:P|(?:var)?p)i|(?:var)?rho|(?:S|(?:var)?s)igma|tau|[Uu]psilon|(?:(?:P|(?:var)?p)|c)hi|[Pp]si|[Oo]mega)$/;
        Visitor.prototype.visit_symbol = new P.Function()
            .when({name: 'pi'},             _ => const_f.bind(undefined, Math.PI))
            .when({name: P('name', greek)}, _ => variable_f.bind(undefined, _.name))
            .when({name: P('name')},        _ => { throw new UnsupportedMathError(_.name); })
            .end;
        Visitor.prototype.visit_bracket = new P.Function()
            // unprocessed integral
            .when(P._, 'int',              P('val'), P('rbkt'), P._, _ => { throw new UnsupportedMathError('int'); })
            // already processed integral
            .when(P._, P('lbkt', /\\int/), P('val'), P('rbkt'), P._, _ => { throw new UnsupportedMathError('int'); })
            .when(P._, '|',                P('val'), '|',       P._, _ => apply1_f.bind(undefined, Math.abs, _.val))
            .when(P._, /\.$/,              P('val'), P('rbkt'), P._, _ => { throw new UnsupportedMathError('hidden bracket'); })
            .when(P._, P('lbkt'),          P('val'), /^\./,     P._, _ => { throw new UnsupportedMathError('hidden bracket'); })
            .when(P._, P('lbkt'),          P('val'), P('rbkt'), P._, _ => _.val)
            .end;
        Visitor.prototype.visit_matrix = function (exp, elements)
        {
            throw new UnsupportedMathError('matrix');
        };

        var toFunction = function (math, options)
        {
            options = options || {};
            var oldyy = parser.yy;
            var errors = [];
            try
            {
                parser.yy = new Simplifier(options);
                var visitor = new Visitor(options);
                var exp = parser.parse(math);
                var func = exp.accept(visitor);
            }
            finally
            {
                parser.yy = oldyy;
            }
            return func;
        };
        toFunction.UnsupportedMathError = UnsupportedMathError;

        return {
            toFunction: toFunction
        };
    };

    // export module depending on what kind of environment we are in
    if (typeof exports === 'object' && typeof module !== 'undefined' && module.exports) {
        // commonjs (e.g. NodeJS)
        var smarttype = require('./smarttype');
        var simplifier = require('./simplifier');
        var P = require('P.js');
        var exp = factory(smarttype, simplifier, P);
        for (var key in exp) {
            exports[key] = exp[key];
        }
    } else if (typeof this.define === 'function' && this.define.amd) {
        // AMD (e.g. requirejs)
        this.define('smarttype/func', ['smarttype', 'smarttype/simplifier', 'Pjs'], function (smarttype, x, Pjs) {
            return {
                toFunction: factory(smarttype, smarttype, Pjs).toFunction
            };
        });
    } else {
        // browser
        this.smarttype = this.smarttype || {};
        this.smarttype.toFunction = factory(this.smarttype, this.smarttype, this.Pjs).toFunction;
    }
}.call(this));
