/* latex es6 */
/* global exports module require */
/* eslint new-cap: 0, no-multi-spaces: 0, key-spacing: 0 */
/**
 * Converts SmartType ( simplifier.js objects ) to LaTeX
 * 
 */
(function () {
    "use strict";

    var factory = function (smarttype, simplifier, P)
    {
        var parser = smarttype.parser;//TODO: factory is called by.
        var Simplifier = simplifier.Simplifier;

        var Visitor = function (options)
        {
            this.options = options;
        };
        Visitor.prototype.visit_addition = function ()
        {
            return Array.prototype.slice.call(arguments, 1).join('');
        };
        Visitor.prototype.visit_addition_sign = function (sign)
        {
            return sign[0] === '\\' ? sign + ' ' : sign;
        };
        Visitor.prototype.visit_multiplication = function ()
        {
            return Array.prototype.slice.call(arguments, 1).join('');
        };
        Visitor.prototype.visit_multiplication_op = function (visible)
        {
            return visible ? "\\cdot " : "";
        };
        var binop_map = {
            '==>': '\\Rightarrow ',
            '<=>': '\\Leftrightarrow ',
            '=^o': '\\circeq ',
            '.=': '\\doteq ',
            'in': '\\in ',
            '<-': '\\leftarrow ',
            '->': '\\rightarrow ',
            'le': '\\leq ',
            '<=': '\\leq ',
            '=<': '\\leq ',
            'ge': '\\geq ',
            '>=': '\\geq ',
            '=>': '\\geq ',
            '~=': '\\cong ',
            'approx': '\\approx ',
            'sub': '\\subset ',
            'sup': '\\supset ',
            'sube': '\\subseteq ',
            'supe': '\\supseteq ',
            'ne': '\\neq ',
            '/==>': '\\nRightarrow ',
            '/<=>': '\\nLeftrightarrow ',
            '/=': '\\neq ',
            '/<': '\\not<',
            '/>': '\\not>',
            '/in': '\\not\\in ',
            '/<-': '\\nleftarrow ',
            '/->': '\\nrightarrow ',
            '/le': '\\nleq ',
            '/<=': '\\nleq ',
            '/=<': '\\nleq ',
            '/ge': '\\ngeq ',
            '/>=': '\\ngeq ',
            '/=>': '\\ngeq ',
            '/~=': '\\ncong ',
            '/approx': '\\not\\approx ',
            '/sub': '\\not\\subset ',
            '/sup': '\\not\\supset ',
            '/sube': '\\not\\subseteq ',
            '/supe': '\\not\\supseteq ',
            'U': '\\cup ',
            'cap': '\\cap '
        };
        Visitor.prototype.visit_binop = new P.Function()
            .when(P._, P('left'), 'root',  P('right'), _ => '\\sqrt[' + _.left + ']{' + _.right + '}')
            .when(P._, P('left'), '_',     P('right'), _ => _.left + '_{' + _.right + '}')
            .when(P._, P('left'), '^',     P('right'), _ => _.left + '^{' + _.right + '}')
            .when(P._, P('left'), '/',     P('right'), _ => '\\frac{' + _.left + '}{' + _.right + '}')
            .when(P._, P('left'), P('op'), P('right'), _ => _.left + (_.op in binop_map ? binop_map[_.op] : _.op) + _.right)
            .end;
        var nonstandard_func = {
            sech: true,
            csch: true,
            arcsech: true,
            arccsch: true,
            arccoth: true
        };
        Visitor.prototype.visit_funcname = function (name)
        {
            if (name in nonstandard_func)
            {
                return '\\mathrm{' + name + '}';
            }
            return '\\' + name + ' ';
        };
        Visitor.prototype.visit_bigop = new P.Function()
            .when(P._, P('op'), null,     null,     _ => '\\' + _.op + ' ')
            .when(P._, P('op'), P('sub'), null,     _ => '\\' + _.op + '_{' + _.sub + '}')
            .when(P._, P('op'), null,     P('sup'), _ => '\\' + _.op + '^{' + _.sup + '}')
            .when(P._, P('op'), P('sub'), P('sup'), _ => '\\' + _.op + '_{' + _.sub + '}' + '^{' + _.sup + '}')
            .end;
        Visitor.prototype.visit_sign = function (exp, sign, value)
        {
            return sign + value;
        };
        Visitor.prototype.visit_func = new P.Function()
            .when(P._, '\\floor ', P('value'), _ => '\\left\\lfloor ' + _.value + '\\right\\rfloor ')
            .when(P._, '\\ceil ',  P('value'), _ => '\\left\\lceil ' + _.value + '\\right\\rceil ')
            .when(P._, '\\sqrt ',  P('value'), _ => '\\sqrt{' + _.value + '}')
            .when(P._, P('name'),  P('value'), _ => _.name + _.value)
            .end;
        Visitor.prototype.visit_postfunc = new P.Function()
            .when(P._, '->',        P('value'), _ => '\\vec{' + _.value + '}')
            .when(P._, '_',         P('value'), _ => '\\bar{' + _.value + '}')
            .when(P._, '^',         P('value'), _ => '\\hat{' + _.value + '}')
            .when(P._, '~',         P('value'), _ => '\\tilde{' + _.value + '}')
            .when(P._, '.',         P('value'), _ => '\\dot{' + _.value + '}')
            .when(P._, 'factorial', P('value'), _ => _.value + '!')
            .when(P._, 'prime',     P.captureRE(/^(\{.*\^\{(?:\\prime)+)\}\}/, 'pre'), _ => _.pre + '\\prime}}') // a hacky way to handle multiple primes
            .when(P._, 'prime',     P('value'), _ => '{' + _.value + '^{\\prime}}')
            .when(P._, P('name'),   P('value'), _ => _.value + _.name)
            .end;
        Visitor.prototype.visit_number = function (num)
        {
            return num.value;
        };
        Visitor.prototype.visit_input = function (input)
        {
            return `\\TextField{${input.name.replace(/(\(|\))/g,'')}}`;//TODO:
        };
        Visitor.prototype.visit_latex = function (latex)
        {
            return latex.latex;
        };
        Visitor.prototype.visit_variable = function (variable)
        {
            return variable.name;
        };
        var greek = /^(?:alpha|beta|[Gg]amma|[Dd]elta|(?:var)?epsilon|(?:z||(?:T|(?:var)?t)h)eta|iota|(?:var)?kappa|[Ll]ambda|[mn]u|[Xx]i|omicron|(?:P|(?:var)?p)i|(?:var)?rho|(?:S|(?:var)?s)igma|tau|[Uu]psilon|(?:(?:P|(?:var)?p)|c)hi|[Pp]si|[Oo]mega)$/;
        Visitor.prototype.visit_symbol = new P.Function()
            .when({name:           null},                    function (_) { return this.options.dispmissing ? '\\Box ' : '{}'; })
            .when({name: P('name', greek)},                  _ => '\\' + _.name + ' ')
            .when({name:           'd.'},                    _ => '\\mathrm{d}')
            .when({name: P.captureRE(/^d(.)$/, 'x')},        _ => '\\mathrm{d}' + _.x)
            .when({name:           'inf'},                   _ => '\\infty ')
            .when({name: P('name', /^Z[-+]$/)},              _ => '\\mathbb{Z}^' + _.name[1])
            .when({name:           '{}' },                   _ => '\\emptyset ')
            .when({name:           '...'},                   _ => '\\dots ')
            .when({name:           'undefined'},             _ => '\\textrm{undefined}')
            .when({name:           'doesnotexist'},          _ => '\\textrm{doesnotexist}')
            .when({name: P('name', /^(?:CC|QQ|NN|RR|ZZ)$/)}, _ => '\\mathbb{' + _.name[0] + '}')
            .when({name: P('name')},                         _ => _.name)
            .end;
        Visitor.prototype.visit_bracket = new P.Function()
            // unprocessed integral
            .when(P._, 'int',              P('val'), P('rbkt'), P._, function (_) { return '\\int ' + _.val + '\\;' + this.visit_symbol({name: _.rbkt}); })
            // already processed integral
            .when(P._, P('lbkt', /\\int/), P('val'), P('rbkt'), P._, function (_) { return _.lbkt + _.val + '\\;' + this.visit_symbol({name: _.rbkt}); })
            .when(P._, /\.$/,              P('val'), /^\./,     P._, _ => _.val)
            .when(P._, /\.$/,              P('val'), '}',       P._, _ => '\\left.' + _.val + '\\right\\}')
            .when(P._, /\.$/,              P('val'), P('rbkt'), P._, _ => '\\left.' + _.val + '\\right' + _.rbkt)
            .when(P._, '{',                P('val'), /^\./,     P._, _ => '\\left\\{' + _.val + '\\right.')
            .when(P._, P('lbkt'),          P('val'), /^\./,     P._, _ => '\\left' + _.lbkt + _.val + '\\right.')
            .when(P._, '{',                P('val'), '}',       P._, _ => '\\left\\{' + _.val + '\\right\\}')
            .when(P._, P('lbkt'),          P('val'), '}',       P._, _ => '\\left' + _.lbkt + _.val + '\\right\\}')
            .when(P._, '{',                P('val'), P('rbkt'), P._, _ => '\\left\\{' + _.val + '\\right' + _.rbkt)
            .when(P._, P('lbkt'),          P('val'), P('rbkt'), P._, _ => '\\left' + _.lbkt + _.val + '\\right' + _.rbkt)
            .end;
        Visitor.prototype.visit_matrix = function (exp, elements)
        {
            var result = "\\begin{matrix} " + elements[0].join(" & ");
            for (var i = 1; i < elements.length; i++)
            {
                var row = elements[i];
                result += " \\\\ " + row.join(" & ");
            }
            result += " \\end{matrix}";
            return result;
        };

        var toLaTeX = function (math, options)
        {
            options = options || {};
            var oldyy = parser.yy;
            var errors = [];
            try
            {
                parser.yy = new Simplifier(options);
                var visitor = new Visitor(options);
                var exp = parser.parse(math);
                var latex = exp.accept(visitor);
            }
            finally
            {
                parser.yy = oldyy;
            }
            return {latex: latex, errors: errors};
        };

        return {
            toLaTeX: toLaTeX
        };
    };

    // export module depending on what kind of environment we are in
    if (typeof exports === 'object' && typeof module !== 'undefined' && module.exports) {
        // commonjs (e.g. NodeJS)
        var smarttype = require('./smarttype');
        var simplifier = require('./simplifier');
        var P = require('P.js');
        var exp = factory(smarttype, simplifier, P);
        for (var key in exp) {
            exports[key] = exp[key];
        }
    } else if (typeof this.define === 'function' && this.define.amd) {
        // AMD (e.g. requirejs)
        this.define('smarttype/latex', ['smarttype', 'smarttype/simplifier', 'Pjs'], function (smarttype, x, Pjs) {
            return {
                toLaTeX: factory(smarttype, smarttype, Pjs).toLaTeX//smarttype=smarttype.js.m4
            };
        });
    } else {
        // browser
        this.smarttype = this.smarttype || {};
        this.smarttype.toLaTeX = factory(this.smarttype, this.smarttype, this.Pjs).toLaTeX;
    }
}.call(this));
